/* Copyright (c) 2018 Griefer@Work                                            *
 *                                                                            *
 * This software is provided 'as-is', without any express or implied          *
 * warranty. In no event will the authors be held liable for any damages      *
 * arising from the use of this software.                                     *
 *                                                                            *
 * Permission is granted to anyone to use this software for any purpose,      *
 * including commercial applications, and to alter it and redistribute it     *
 * freely, subject to the following restrictions:                             *
 *                                                                            *
 * 1. The origin of this software must not be misrepresented; you must not    *
 *    claim that you wrote the original software. If you use this software    *
 *    in a product, an acknowledgement in the product documentation would be  *
 *    appreciated but is not required.                                        *
 * 2. Altered source versions must be plainly marked as such, and must not be *
 *    misrepresented as being the original software.                          *
 * 3. This notice may not be removed or altered from any source distribution. *
 */
#ifndef _SYS_GENERIC_UTSNAME_H
#define _SYS_GENERIC_UTSNAME_H 1
#define _SYS_UTSNAME_H 1

#include <__stdinc.h>
#include <features.h>
#include <bits/utsname.h>
#ifdef __USE_KOS
#if __KOS_VERSION__ < 300
#include <kos/ksym.h>
#endif
#endif /* __USE_KOS */

__SYSDECL_BEGIN

#ifdef __COMPILER_HAVE_PRAGMA_PUSHMACRO
#pragma push_macro("sysname")
#pragma push_macro("nodename")
#pragma push_macro("release")
#pragma push_macro("version")
#pragma push_macro("machine")
#pragma push_macro("domainname")
#endif

#undef sysname
#undef nodename
#undef release
#undef version
#undef machine
#undef domainname

#ifndef _UTSNAME_SYSNAME_LENGTH
#define _UTSNAME_SYSNAME_LENGTH _UTSNAME_LENGTH
#endif
#ifndef _UTSNAME_NODENAME_LENGTH
#define _UTSNAME_NODENAME_LENGTH _UTSNAME_LENGTH
#endif
#ifndef _UTSNAME_RELEASE_LENGTH
#define _UTSNAME_RELEASE_LENGTH _UTSNAME_LENGTH
#endif
#ifndef _UTSNAME_VERSION_LENGTH
#define _UTSNAME_VERSION_LENGTH _UTSNAME_LENGTH
#endif
#ifndef _UTSNAME_MACHINE_LENGTH
#define _UTSNAME_MACHINE_LENGTH _UTSNAME_LENGTH
#endif

struct utsname {
    char   sysname[_UTSNAME_SYSNAME_LENGTH];   /* Name of the implementation of the operating system. */
    char   nodename[_UTSNAME_NODENAME_LENGTH]; /* Name of this node on the network. */
    char   release[_UTSNAME_RELEASE_LENGTH];   /* Current release level of this implementation. */
    char   version[_UTSNAME_VERSION_LENGTH];   /* Current version level of this release. */
    char   machine[_UTSNAME_MACHINE_LENGTH];   /* Name of the hardware type the system is running on. */
#if (_UTSNAME_DOMAIN_LENGTH+0) != 0
#ifdef __USE_GNU
    char   domainname[_UTSNAME_DOMAIN_LENGTH]; /* Name of the domain of this node on the network. */
#else
    char __domainname[_UTSNAME_DOMAIN_LENGTH]; /* Name of the domain of this node on the network. */
#endif
#endif
};

#if __KOS_VERSION__ < 300
#ifdef __USE_KOS
/* System-global kernel variable (May be accessed directly). */
__PUBDEF struct utsname const active_uname __KSYM(uname);
#endif /* __USE_KOS */
#endif /* __KOS_VERSION__ < 300 */

#ifdef __USE_MISC
#define SYS_NMLN  _UTSNAME_LENGTH
#endif /* __USE_MISC */

#ifndef __KERNEL__
__REDIRECT_EXCEPT_XVOID(__LIBC,,int,__LIBCCALL,uname,(struct utsname *__name),(__name))
#ifdef __USE_EXCEPT
__LIBC void (__LIBCCALL Xuname)(struct utsname *__name);
#endif
#endif /* !__KERNEL__ */

#ifdef __COMPILER_HAVE_PRAGMA_PUSHMACRO
#pragma pop_macro("domainname")
#pragma pop_macro("machine")
#pragma pop_macro("version")
#pragma pop_macro("release")
#pragma pop_macro("nodename")
#pragma pop_macro("sysname")
#endif

__SYSDECL_END

#endif /* !_SYS_GENERIC_UTSNAME_H */
