/* Copyright (c) 2018 Griefer@Work                                            *
 *                                                                            *
 * This software is provided 'as-is', without any express or implied          *
 * warranty. In no event will the authors be held liable for any damages      *
 * arising from the use of this software.                                     *
 *                                                                            *
 * Permission is granted to anyone to use this software for any purpose,      *
 * including commercial applications, and to alter it and redistribute it     *
 * freely, subject to the following restrictions:                             *
 *                                                                            *
 * 1. The origin of this software must not be misrepresented; you must not    *
 *    claim that you wrote the original software. If you use this software    *
 *    in a product, an acknowledgement in the product documentation would be  *
 *    appreciated but is not required.                                        *
 * 2. Altered source versions must be plainly marked as such, and must not be *
 *    misrepresented as being the original software.                          *
 * 3. This notice may not be removed or altered from any source distribution. *
 */
#ifndef GUARD_LIBS_LIBC_FORMAT_WRAPPER_C_INL
#define GUARD_LIBS_LIBC_FORMAT_WRAPPER_C_INL 1
#define _UTF_SOURCE 1

#include "libc.h"
#include "format.h"

#include <hybrid/compiler.h>
#include <format-printer.h>
#include <uchar.h>
#include <assert.h>
#include <except.h>

DECL_BEGIN

#define PRINTF_UNPACK(...)   __VA_ARGS__
#define PRINTF_CALL(x,args)  x args

#define DEFINE_PRINTF(Treturn,printf,args,vprintf,params,before_va_start) \
INTERN Treturn (ATTR_CDECL printf) args { \
  Treturn COMPILER_IGNORE_UNINITIALIZED(result); \
  va_list __EXCEPTVAR_VALIST varargs; \
  va_start(varargs,before_va_start); \
  __TRY_VALIST { \
    result = PRINTF_CALL(vprintf,(PRINTF_UNPACK params,varargs)); \
  } __FINALLY_VALIST { \
    va_end(varargs); \
  } \
  return result; \
}

DEFINE_PRINTF(ssize_t,libc_format_w16scanf,(pw16formatgetc scanner, pw16formatungetc returnch, void *closure, char16_t const *__restrict format, ...),libc_format_vw16scanf,(scanner,returnch,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w16scanf,(pw16formatgetc scanner, pw16formatungetc returnch, void *closure, char16_t const *__restrict format, ...),libd_format_vw16scanf,(scanner,returnch,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w32scanf,(pw32formatgetc scanner, pw32formatungetc returnch, void *closure, char32_t const *__restrict format, ...),libc_format_vw32scanf,(scanner,returnch,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w32scanf,(pw32formatgetc scanner, pw32formatungetc returnch, void *closure, char32_t const *__restrict format, ...),libd_format_vw32scanf,(scanner,returnch,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w16scanf_l,(pw16formatgetc scanner, pw16formatungetc returnch, void *closure, char16_t const *__restrict format, locale_t locale, ...),libc_format_vw16scanf_l,(scanner,returnch,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w16scanf_l,(pw16formatgetc scanner, pw16formatungetc returnch, void *closure, char16_t const *__restrict format, locale_t locale, ...),libd_format_vw16scanf_l,(scanner,returnch,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w32scanf_l,(pw32formatgetc scanner, pw32formatungetc returnch, void *closure, char32_t const *__restrict format, locale_t locale, ...),libc_format_vw32scanf_l,(scanner,returnch,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w32scanf_l,(pw32formatgetc scanner, pw32formatungetc returnch, void *closure, char32_t const *__restrict format, locale_t locale, ...),libd_format_vw32scanf_l,(scanner,returnch,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w16printf,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libc_format_vw16printf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w16printf,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libd_format_vw16printf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w32printf,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libc_format_vw32printf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w32printf,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libd_format_vw32printf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w16printf_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libc_format_vw16printf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w16printf_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libd_format_vw16printf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w32printf_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libc_format_vw32printf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w32printf_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libd_format_vw32printf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_printf_p,(pformatprinter printer, void *closure, char const *__restrict format, ...),libc_format_vprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_printf_p,(pformatprinter printer, void *closure, char const *__restrict format, ...),libd_format_vprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w16printf_p,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libc_format_vw16printf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w16printf_p,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libd_format_vw16printf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w32printf_p,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libc_format_vw32printf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w32printf_p,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libd_format_vw32printf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_printf_p_l,(pformatprinter printer, void *closure, char const *__restrict format, locale_t locale, ...),libc_format_vprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_printf_p_l,(pformatprinter printer, void *closure, char const *__restrict format, locale_t locale, ...),libd_format_vprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w16printf_p_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libc_format_vw16printf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w16printf_p_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libd_format_vw16printf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w32printf_p_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libc_format_vw32printf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w32printf_p_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libd_format_vw32printf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_bprintf,(pformatprinter printer, void *closure, char const *__restrict format, ...),libc_format_vbprintf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_bprintf,(pformatprinter printer, void *closure, char const *__restrict format, ...),libd_format_vbprintf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w16bprintf,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libc_format_vw16bprintf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w16bprintf,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libd_format_vw16bprintf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w32bprintf,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libc_format_vw32bprintf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w32bprintf,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libd_format_vw32bprintf,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_bprintf_p,(pformatprinter printer, void *closure, char const *__restrict format, ...),libc_format_vbprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_bprintf_p,(pformatprinter printer, void *closure, char const *__restrict format, ...),libd_format_vbprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w16bprintf_p,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libc_format_vw16bprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w16bprintf_p,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, ...),libd_format_vw16bprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_w32bprintf_p,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libc_format_vw32bprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libd_format_w32bprintf_p,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, ...),libd_format_vw32bprintf_p,(printer,closure,format),format)
DEFINE_PRINTF(ssize_t,libc_format_bprintf_l,(pformatprinter printer, void *closure, char const *__restrict format, locale_t locale, ...),libc_format_vbprintf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_bprintf_l,(pformatprinter printer, void *closure, char const *__restrict format, locale_t locale, ...),libd_format_vbprintf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w16bprintf_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libc_format_vw16bprintf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w16bprintf_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libd_format_vw16bprintf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w32bprintf_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libc_format_vw32bprintf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w32bprintf_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libd_format_vw32bprintf_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_bprintf_p_l,(pformatprinter printer, void *closure, char const *__restrict format, locale_t locale, ...),libc_format_vbprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_bprintf_p_l,(pformatprinter printer, void *closure, char const *__restrict format, locale_t locale, ...),libd_format_vbprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w16bprintf_p_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libc_format_vw16bprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w16bprintf_p_l,(pw16formatprinter printer, void *closure, char16_t const *__restrict format, locale_t locale, ...),libd_format_vw16bprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_format_w32bprintf_p_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libc_format_vw32bprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_format_w32bprintf_p_l,(pw32formatprinter printer, void *closure, char32_t const *__restrict format, locale_t locale, ...),libd_format_vw32bprintf_p_l,(printer,closure,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sprintf,(char *__restrict buf, char const *__restrict format, ...),libc_vsprintf,(buf,format),format)
DEFINE_PRINTF(size_t,libd_sprintf,(char *__restrict buf, char const *__restrict format, ...),libd_vsprintf,(buf,format),format)
DEFINE_PRINTF(size_t,libc_sw16printf,(char16_t *__restrict buf, char16_t const *__restrict format, ...),libc_vsw16printf,(buf,format),format)
DEFINE_PRINTF(size_t,libd_sw16printf,(char16_t *__restrict buf, char16_t const *__restrict format, ...),libd_vsw16printf,(buf,format),format)
DEFINE_PRINTF(size_t,libc_sw32printf,(char32_t *__restrict buf, char32_t const *__restrict format, ...),libc_vsw32printf,(buf,format),format)
DEFINE_PRINTF(size_t,libd_sw32printf,(char32_t *__restrict buf, char32_t const *__restrict format, ...),libd_vsw32printf,(buf,format),format)
DEFINE_PRINTF(size_t,libc_sprintf_p,(char *__restrict buf, char const *__restrict format, ...),libc_vsprintf_p,(buf,format),format)
DEFINE_PRINTF(size_t,libd_sprintf_p,(char *__restrict buf, char const *__restrict format, ...),libd_vsprintf_p,(buf,format),format)
DEFINE_PRINTF(size_t,libc_sw16printf_p,(char16_t *__restrict buf, char16_t const *__restrict format, ...),libc_vsw16printf_p,(buf,format),format)
DEFINE_PRINTF(size_t,libd_sw16printf_p,(char16_t *__restrict buf, char16_t const *__restrict format, ...),libd_vsw16printf_p,(buf,format),format)
DEFINE_PRINTF(size_t,libc_sw32printf_p,(char32_t *__restrict buf, char32_t const *__restrict format, ...),libc_vsw32printf_p,(buf,format),format)
DEFINE_PRINTF(size_t,libd_sw32printf_p,(char32_t *__restrict buf, char32_t const *__restrict format, ...),libd_vsw32printf_p,(buf,format),format)
DEFINE_PRINTF(size_t,libc_sprintf_l,(char *__restrict buf, char const *__restrict format, locale_t locale, ...),libc_vsprintf_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sprintf_l,(char *__restrict buf, char const *__restrict format, locale_t locale, ...),libd_vsprintf_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sw16printf_l,(char16_t *__restrict buf, char16_t const *__restrict format, locale_t locale, ...),libc_vsw16printf_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sw16printf_l,(char16_t *__restrict buf, char16_t const *__restrict format, locale_t locale, ...),libd_vsw16printf_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sw32printf_l,(char32_t *__restrict buf, char32_t const *__restrict format, locale_t locale, ...),libc_vsw32printf_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sw32printf_l,(char32_t *__restrict buf, char32_t const *__restrict format, locale_t locale, ...),libd_vsw32printf_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sprintf_p_l,(char *__restrict buf, char const *__restrict format, locale_t locale, ...),libc_vsprintf_p_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sprintf_p_l,(char *__restrict buf, char const *__restrict format, locale_t locale, ...),libd_vsprintf_p_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sw16printf_p_l,(char16_t *__restrict buf, char16_t const *__restrict format, locale_t locale, ...),libc_vsw16printf_p_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sw16printf_p_l,(char16_t *__restrict buf, char16_t const *__restrict format, locale_t locale, ...),libd_vsw16printf_p_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sw32printf_p_l,(char32_t *__restrict buf, char32_t const *__restrict format, locale_t locale, ...),libc_vsw32printf_p_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sw32printf_p_l,(char32_t *__restrict buf, char32_t const *__restrict format, locale_t locale, ...),libd_vsw32printf_p_l,(buf,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snprintf,(char *__restrict buf, size_t buflen, char const *__restrict format, ...),libc_vsnprintf,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snprintf,(char *__restrict buf, size_t buflen, char const *__restrict format, ...),libd_vsnprintf,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snw16printf,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, ...),libc_vsnw16printf,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snw16printf,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, ...),libd_vsnw16printf,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snw32printf,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, ...),libc_vsnw32printf,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snw32printf,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, ...),libd_vsnw32printf,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snprintf_l,(char *__restrict buf, size_t buflen, char const *__restrict format, locale_t locale, ...),libc_vsnprintf_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snprintf_l,(char *__restrict buf, size_t buflen, char const *__restrict format, locale_t locale, ...),libd_vsnprintf_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw16printf_l,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, locale_t locale, ...),libc_vsnw16printf_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw16printf_l,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, locale_t locale, ...),libd_vsnw16printf_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw32printf_l,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, locale_t locale, ...),libc_vsnw32printf_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw32printf_l,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, locale_t locale, ...),libd_vsnw32printf_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snprintf_p,(char *__restrict buf, size_t buflen, char const *__restrict format, ...),libc_vsnprintf_p,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snprintf_p,(char *__restrict buf, size_t buflen, char const *__restrict format, ...),libd_vsnprintf_p,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snw16printf_p,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, ...),libc_vsnw16printf_p,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snw16printf_p,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, ...),libd_vsnw16printf_p,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snw32printf_p,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, ...),libc_vsnw32printf_p,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snw32printf_p,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, ...),libd_vsnw32printf_p,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snprintf_p_l,(char *__restrict buf, size_t buflen, char const *__restrict format, locale_t locale, ...),libc_vsnprintf_p_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snprintf_p_l,(char *__restrict buf, size_t buflen, char const *__restrict format, locale_t locale, ...),libd_vsnprintf_p_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw16printf_p_l,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, locale_t locale, ...),libc_vsnw16printf_p_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw16printf_p_l,(char16_t *__restrict buf, size_t buflen, char16_t const *__restrict format, locale_t locale, ...),libd_vsnw16printf_p_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw32printf_p_l,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, locale_t locale, ...),libc_vsnw32printf_p_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw32printf_p_l,(char32_t *__restrict buf, size_t buflen, char32_t const *__restrict format, locale_t locale, ...),libd_vsnw32printf_p_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snprintf_c,(char *buf, size_t buflen, char const *__restrict format, ...),libc_vsnprintf_c,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snprintf_c,(char *buf, size_t buflen, char const *__restrict format, ...),libd_vsnprintf_c,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snprintf_c_l,(char *buf, size_t buflen, char const *__restrict format, locale_t locale, ...),libc_vsnprintf_c_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snprintf_c_l,(char *buf, size_t buflen, char const *__restrict format, locale_t locale, ...),libd_vsnprintf_c_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw16printf_c,(char16_t *buf, size_t buflen, char16_t const *__restrict format, ...),libc_vsnw16printf_c,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snw16printf_c,(char16_t *buf, size_t buflen, char16_t const *__restrict format, ...),libd_vsnw16printf_c,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snw16printf_c_l,(char16_t *buf, size_t buflen, char16_t const *__restrict format, locale_t locale, ...),libc_vsnw16printf_c_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw16printf_c_l,(char16_t *buf, size_t buflen, char16_t const *__restrict format, locale_t locale, ...),libd_vsnw16printf_c_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw32printf_c,(char32_t *buf, size_t buflen, char32_t const *__restrict format, ...),libc_vsnw32printf_c,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libd_snw32printf_c,(char32_t *buf, size_t buflen, char32_t const *__restrict format, ...),libd_vsnw32printf_c,(buf,buflen,format),format)
DEFINE_PRINTF(size_t,libc_snw32printf_c_l,(char32_t *buf, size_t buflen, char32_t const *__restrict format, locale_t locale, ...),libc_vsnw32printf_c_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw32printf_c_l,(char32_t *buf, size_t buflen, char32_t const *__restrict format, locale_t locale, ...),libd_vsnw32printf_c_l,(buf,buflen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_scprintf,(char const *__restrict format, ...),libc_vscprintf,(format),format)
DEFINE_PRINTF(size_t,libd_scprintf,(char const *__restrict format, ...),libd_vscprintf,(format),format)
DEFINE_PRINTF(size_t,libc_scw16printf,(char16_t const *__restrict format, ...),libc_vscw16printf,(format),format)
DEFINE_PRINTF(size_t,libd_scw16printf,(char16_t const *__restrict format, ...),libd_vscw16printf,(format),format)
DEFINE_PRINTF(size_t,libc_scw32printf,(char32_t const *__restrict format, ...),libc_vscw32printf,(format),format)
DEFINE_PRINTF(size_t,libd_scw32printf,(char32_t const *__restrict format, ...),libd_vscw32printf,(format),format)
DEFINE_PRINTF(size_t,libc_scprintf_p,(char const *__restrict format, ...),libc_vscprintf_p,(format),format)
DEFINE_PRINTF(size_t,libd_scprintf_p,(char const *__restrict format, ...),libd_vscprintf_p,(format),format)
DEFINE_PRINTF(size_t,libc_scw16printf_p,(char16_t const *__restrict format, ...),libc_vscw16printf_p,(format),format)
DEFINE_PRINTF(size_t,libd_scw16printf_p,(char16_t const *__restrict format, ...),libd_vscw16printf_p,(format),format)
DEFINE_PRINTF(size_t,libc_scw32printf_p,(char32_t const *__restrict format, ...),libc_vscw32printf_p,(format),format)
DEFINE_PRINTF(size_t,libd_scw32printf_p,(char32_t const *__restrict format, ...),libd_vscw32printf_p,(format),format)
DEFINE_PRINTF(size_t,libc_scprintf_l,(char const *__restrict format, locale_t locale, ...),libc_vscprintf_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libd_scprintf_l,(char const *__restrict format, locale_t locale, ...),libd_vscprintf_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libc_scw16printf_l,(char16_t const *__restrict format, locale_t locale, ...),libc_vscw16printf_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libd_scw16printf_l,(char16_t const *__restrict format, locale_t locale, ...),libd_vscw16printf_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libc_scw32printf_l,(char32_t const *__restrict format, locale_t locale, ...),libc_vscw32printf_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libd_scw32printf_l,(char32_t const *__restrict format, locale_t locale, ...),libd_vscw32printf_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libc_scprintf_p_l,(char const *__restrict format, locale_t locale, ...),libc_vscprintf_p_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libd_scprintf_p_l,(char const *__restrict format, locale_t locale, ...),libd_vscprintf_p_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libc_scw16printf_p_l,(char16_t const *__restrict format, locale_t locale, ...),libc_vscw16printf_p_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libd_scw16printf_p_l,(char16_t const *__restrict format, locale_t locale, ...),libd_vscw16printf_p_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libc_scw32printf_p_l,(char32_t const *__restrict format, locale_t locale, ...),libc_vscw32printf_p_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libd_scw32printf_p_l,(char32_t const *__restrict format, locale_t locale, ...),libd_vscw32printf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_dprintf,(fd_t fd, char const *__restrict format, ...),libc_vdprintf,(fd,format),format)
DEFINE_PRINTF(ssize_t,libd_dprintf,(fd_t fd, char const *__restrict format, ...),libd_vdprintf,(fd,format),format)
DEFINE_PRINTF(ssize_t,libc_w16dprintf,(fd_t fd, char16_t const *__restrict format, ...),libc_vw16dprintf,(fd,format),format)
DEFINE_PRINTF(ssize_t,libd_w16dprintf,(fd_t fd, char16_t const *__restrict format, ...),libd_vw16dprintf,(fd,format),format)
DEFINE_PRINTF(ssize_t,libc_w32dprintf,(fd_t fd, char32_t const *__restrict format, ...),libc_vw32dprintf,(fd,format),format)
DEFINE_PRINTF(ssize_t,libd_w32dprintf,(fd_t fd, char32_t const *__restrict format, ...),libd_vw32dprintf,(fd,format),format)
DEFINE_PRINTF(ssize_t,libc_dprintf_p,(fd_t fd, char const *__restrict format, ...),libc_vdprintf_p,(fd,format),format)
DEFINE_PRINTF(ssize_t,libd_dprintf_p,(fd_t fd, char const *__restrict format, ...),libd_vdprintf_p,(fd,format),format)
DEFINE_PRINTF(ssize_t,libc_w16dprintf_p,(fd_t fd, char16_t const *__restrict format, ...),libc_vw16dprintf_p,(fd,format),format)
DEFINE_PRINTF(ssize_t,libd_w16dprintf_p,(fd_t fd, char16_t const *__restrict format, ...),libd_vw16dprintf_p,(fd,format),format)
DEFINE_PRINTF(ssize_t,libc_w32dprintf_p,(fd_t fd, char32_t const *__restrict format, ...),libc_vw32dprintf_p,(fd,format),format)
DEFINE_PRINTF(ssize_t,libd_w32dprintf_p,(fd_t fd, char32_t const *__restrict format, ...),libd_vw32dprintf_p,(fd,format),format)
DEFINE_PRINTF(ssize_t,libc_dprintf_l,(fd_t fd, char const *__restrict format, locale_t locale, ...),libc_vdprintf_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_dprintf_l,(fd_t fd, char const *__restrict format, locale_t locale, ...),libd_vdprintf_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w16dprintf_l,(fd_t fd, char16_t const *__restrict format, locale_t locale, ...),libc_vw16dprintf_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w16dprintf_l,(fd_t fd, char16_t const *__restrict format, locale_t locale, ...),libd_vw16dprintf_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w32dprintf_l,(fd_t fd, char32_t const *__restrict format, locale_t locale, ...),libc_vw32dprintf_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w32dprintf_l,(fd_t fd, char32_t const *__restrict format, locale_t locale, ...),libd_vw32dprintf_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_dprintf_p_l,(fd_t fd, char const *__restrict format, locale_t locale, ...),libc_vdprintf_p_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_dprintf_p_l,(fd_t fd, char const *__restrict format, locale_t locale, ...),libd_vdprintf_p_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w16dprintf_p_l,(fd_t fd, char16_t const *__restrict format, locale_t locale, ...),libc_vw16dprintf_p_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w16dprintf_p_l,(fd_t fd, char16_t const *__restrict format, locale_t locale, ...),libd_vw16dprintf_p_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w32dprintf_p_l,(fd_t fd, char32_t const *__restrict format, locale_t locale, ...),libc_vw32dprintf_p_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w32dprintf_p_l,(fd_t fd, char32_t const *__restrict format, locale_t locale, ...),libd_vw32dprintf_p_l,(fd,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_asprintf,(char **__restrict ptr, char const *__restrict format, ...),libc_vasprintf,(ptr,format),format)
DEFINE_PRINTF(ssize_t,libd_asprintf,(char **__restrict ptr, char const *__restrict format, ...),libd_vasprintf,(ptr,format),format)
DEFINE_PRINTF(ssize_t,libc_asprintf_p,(char **__restrict ptr, char const *__restrict format, ...),libc_vasprintf_p,(ptr,format),format)
DEFINE_PRINTF(ssize_t,libd_asprintf_p,(char **__restrict ptr, char const *__restrict format, ...),libd_vasprintf_p,(ptr,format),format)
DEFINE_PRINTF(ssize_t,libc_asprintf_l,(char **__restrict ptr, char const *__restrict format, locale_t locale, ...),libc_vasprintf_l,(ptr,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_asprintf_l,(char **__restrict ptr, char const *__restrict format, locale_t locale, ...),libd_vasprintf_l,(ptr,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_asprintf_p_l,(char **__restrict ptr, char const *__restrict format, locale_t locale, ...),libc_vasprintf_p_l,(ptr,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_asprintf_p_l,(char **__restrict ptr, char const *__restrict format, locale_t locale, ...),libd_vasprintf_p_l,(ptr,format,locale),locale)
DEFINE_PRINTF(char *,libc_strdupf,(char const *__restrict format, ...),libc_vstrdupf,(format),format)
DEFINE_PRINTF(char *,libd_strdupf,(char const *__restrict format, ...),libd_vstrdupf,(format),format)
DEFINE_PRINTF(char *,libc_strdupf_p,(char const *__restrict format, ...),libc_vstrdupf_p,(format),format)
DEFINE_PRINTF(char *,libd_strdupf_p,(char const *__restrict format, ...),libd_vstrdupf_p,(format),format)
DEFINE_PRINTF(char *,libc_strdupf_l,(char const *__restrict format, locale_t locale, ...),libc_vstrdupf_l,(format,locale),locale)
DEFINE_PRINTF(char *,libd_strdupf_l,(char const *__restrict format, locale_t locale, ...),libd_vstrdupf_l,(format,locale),locale)
DEFINE_PRINTF(char *,libc_strdupf_p_l,(char const *__restrict format, locale_t locale, ...),libc_vstrdupf_p_l,(format,locale),locale)
DEFINE_PRINTF(char *,libd_strdupf_p_l,(char const *__restrict format, locale_t locale, ...),libd_vstrdupf_p_l,(format,locale),locale)
DEFINE_PRINTF(char *,libc_Xstrdupf,(char const *__restrict format, ...),libc_Xvstrdupf,(format),format)
DEFINE_PRINTF(char *,libd_Xstrdupf,(char const *__restrict format, ...),libd_Xvstrdupf,(format),format)
DEFINE_PRINTF(char *,libc_Xstrdupf_p,(char const *__restrict format, ...),libc_Xvstrdupf_p,(format),format)
DEFINE_PRINTF(char *,libd_Xstrdupf_p,(char const *__restrict format, ...),libd_Xvstrdupf_p,(format),format)
DEFINE_PRINTF(char *,libc_Xstrdupf_l,(char const *__restrict format, locale_t locale, ...),libc_Xvstrdupf_l,(format,locale),locale)
DEFINE_PRINTF(char *,libd_Xstrdupf_l,(char const *__restrict format, locale_t locale, ...),libd_Xvstrdupf_l,(format,locale),locale)
DEFINE_PRINTF(char *,libc_Xstrdupf_p_l,(char const *__restrict format, locale_t locale, ...),libc_Xvstrdupf_p_l,(format,locale),locale)
DEFINE_PRINTF(char *,libd_Xstrdupf_p_l,(char const *__restrict format, locale_t locale, ...),libd_Xvstrdupf_p_l,(format,locale),locale)
DEFINE_PRINTF(size_t,libc_sscanf,(char const *__restrict src, char const *__restrict format, ...),libc_vsscanf,(src,format),format)
DEFINE_PRINTF(size_t,libd_sscanf,(char const *__restrict src, char const *__restrict format, ...),libd_vsscanf,(src,format),format)
DEFINE_PRINTF(size_t,libc_sw16scanf,(char16_t const *__restrict src, char16_t const *__restrict format, ...),libc_vsw16scanf,(src,format),format)
DEFINE_PRINTF(size_t,libd_sw16scanf,(char16_t const *__restrict src, char16_t const *__restrict format, ...),libd_vsw16scanf,(src,format),format)
DEFINE_PRINTF(size_t,libc_sw32scanf,(char32_t const *__restrict src, char32_t const *__restrict format, ...),libc_vsw32scanf,(src,format),format)
DEFINE_PRINTF(size_t,libd_sw32scanf,(char32_t const *__restrict src, char32_t const *__restrict format, ...),libd_vsw32scanf,(src,format),format)
DEFINE_PRINTF(size_t,libc_sscanf_l,(char const *__restrict src, char const *__restrict format, locale_t locale, ...),libc_vsscanf_l,(src,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sscanf_l,(char const *__restrict src, char const *__restrict format, locale_t locale, ...),libd_vsscanf_l,(src,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sw16scanf_l,(char16_t const *__restrict src, char16_t const *__restrict format, locale_t locale, ...),libc_vsw16scanf_l,(src,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sw16scanf_l,(char16_t const *__restrict src, char16_t const *__restrict format, locale_t locale, ...),libd_vsw16scanf_l,(src,format,locale),locale)
DEFINE_PRINTF(size_t,libc_sw32scanf_l,(char32_t const *__restrict src, char32_t const *__restrict format, locale_t locale, ...),libc_vsw32scanf_l,(src,format,locale),locale)
DEFINE_PRINTF(size_t,libd_sw32scanf_l,(char32_t const *__restrict src, char32_t const *__restrict format, locale_t locale, ...),libd_vsw32scanf_l,(src,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snscanf,(char const *__restrict src, size_t srclen, char const *__restrict format, ...),libc_vsnscanf,(src,srclen,format),format)
DEFINE_PRINTF(size_t,libd_snscanf,(char const *__restrict src, size_t srclen, char const *__restrict format, ...),libd_vsnscanf,(src,srclen,format),format)
DEFINE_PRINTF(size_t,libc_snw16scanf,(char16_t const *__restrict src, size_t srclen, char16_t const *__restrict format, ...),libc_vsnw16scanf,(src,srclen,format),format)
DEFINE_PRINTF(size_t,libd_snw16scanf,(char16_t const *__restrict src, size_t srclen, char16_t const *__restrict format, ...),libd_vsnw16scanf,(src,srclen,format),format)
DEFINE_PRINTF(size_t,libc_snw32scanf,(char32_t const *__restrict src, size_t srclen, char32_t const *__restrict format, ...),libc_vsnw32scanf,(src,srclen,format),format)
DEFINE_PRINTF(size_t,libd_snw32scanf,(char32_t const *__restrict src, size_t srclen, char32_t const *__restrict format, ...),libd_vsnw32scanf,(src,srclen,format),format)
DEFINE_PRINTF(size_t,libc_snscanf_l,(char const *__restrict src, size_t srclen, char const *__restrict format, locale_t locale, ...),libc_vsnscanf_l,(src,srclen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snscanf_l,(char const *__restrict src, size_t srclen, char const *__restrict format, locale_t locale, ...),libd_vsnscanf_l,(src,srclen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw16scanf_l,(char16_t const *__restrict src, size_t srclen, char16_t const *__restrict format, locale_t locale, ...),libc_vsnw16scanf_l,(src,srclen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw16scanf_l,(char16_t const *__restrict src, size_t srclen, char16_t const *__restrict format, locale_t locale, ...),libd_vsnw16scanf_l,(src,srclen,format,locale),locale)
DEFINE_PRINTF(size_t,libc_snw32scanf_l,(char32_t const *__restrict src, size_t srclen, char32_t const *__restrict format, locale_t locale, ...),libc_vsnw32scanf_l,(src,srclen,format,locale),locale)
DEFINE_PRINTF(size_t,libd_snw32scanf_l,(char32_t const *__restrict src, size_t srclen, char32_t const *__restrict format, locale_t locale, ...),libd_vsnw32scanf_l,(src,srclen,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_printf,(char const *__restrict format, ...),libc_vprintf,(format),format)
DEFINE_PRINTF(ssize_t,libd_printf,(char const *__restrict format, ...),libd_vprintf,(format),format)
DEFINE_PRINTF(ssize_t,libc_w16printf,(char16_t const *__restrict format, ...),libc_vw16printf,(format),format)
DEFINE_PRINTF(ssize_t,libd_w16printf,(char16_t const *__restrict format, ...),libd_vw16printf,(format),format)
DEFINE_PRINTF(ssize_t,libc_w32printf,(char32_t const *__restrict format, ...),libc_vw32printf,(format),format)
DEFINE_PRINTF(ssize_t,libd_w32printf,(char32_t const *__restrict format, ...),libd_vw32printf,(format),format)
DEFINE_PRINTF(ssize_t,libc_printf_p,(char const *__restrict format, ...),libc_vprintf_p,(format),format)
DEFINE_PRINTF(ssize_t,libd_printf_p,(char const *__restrict format, ...),libd_vprintf_p,(format),format)
DEFINE_PRINTF(ssize_t,libc_w16printf_p,(char16_t const *__restrict format, ...),libc_vw16printf_p,(format),format)
DEFINE_PRINTF(ssize_t,libd_w16printf_p,(char16_t const *__restrict format, ...),libd_vw16printf_p,(format),format)
DEFINE_PRINTF(ssize_t,libc_w32printf_p,(char32_t const *__restrict format, ...),libc_vw32printf_p,(format),format)
DEFINE_PRINTF(ssize_t,libd_w32printf_p,(char32_t const *__restrict format, ...),libd_vw32printf_p,(format),format)
DEFINE_PRINTF(ssize_t,libc_printf_l,(char const *__restrict format, locale_t locale, ...),libc_vprintf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_printf_l,(char const *__restrict format, locale_t locale, ...),libd_vprintf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w16printf_l,(char16_t const *__restrict format, locale_t locale, ...),libc_vw16printf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w16printf_l,(char16_t const *__restrict format, locale_t locale, ...),libd_vw16printf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w32printf_l,(char32_t const *__restrict format, locale_t locale, ...),libc_vw32printf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w32printf_l,(char32_t const *__restrict format, locale_t locale, ...),libd_vw32printf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_printf_p_l,(char const *__restrict format, locale_t locale, ...),libc_vprintf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_printf_p_l,(char const *__restrict format, locale_t locale, ...),libd_vprintf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w16printf_p_l,(char16_t const *__restrict format, locale_t locale, ...),libc_vw16printf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w16printf_p_l,(char16_t const *__restrict format, locale_t locale, ...),libd_vw16printf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w32printf_p_l,(char32_t const *__restrict format, locale_t locale, ...),libc_vw32printf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w32printf_p_l,(char32_t const *__restrict format, locale_t locale, ...),libd_vw32printf_p_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fprintf,(FILE *__restrict self, char const *__restrict format, ...),libc_vfprintf,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fprintf,(FILE *__restrict self, char const *__restrict format, ...),libd_vfprintf,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fw16printf,(FILE *__restrict self, char16_t const *__restrict format, ...),libc_vfw16printf,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fw16printf,(FILE *__restrict self, char16_t const *__restrict format, ...),libd_vfw16printf,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fw32printf,(FILE *__restrict self, char32_t const *__restrict format, ...),libc_vfw32printf,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fw32printf,(FILE *__restrict self, char32_t const *__restrict format, ...),libd_vfw32printf,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fprintf_p,(FILE *__restrict self, char const *__restrict format, ...),libc_vfprintf_p,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fprintf_p,(FILE *__restrict self, char const *__restrict format, ...),libd_vfprintf_p,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fw16printf_p,(FILE *__restrict self, char16_t const *__restrict format, ...),libc_vfw16printf_p,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fw16printf_p,(FILE *__restrict self, char16_t const *__restrict format, ...),libd_vfw16printf_p,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fw32printf_p,(FILE *__restrict self, char32_t const *__restrict format, ...),libc_vfw32printf_p,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fw32printf_p,(FILE *__restrict self, char32_t const *__restrict format, ...),libd_vfw32printf_p,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fprintf_l,(FILE *__restrict self, char const *__restrict format, locale_t locale, ...),libc_vfprintf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fprintf_l,(FILE *__restrict self, char const *__restrict format, locale_t locale, ...),libd_vfprintf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fw16printf_l,(FILE *__restrict self, char16_t const *__restrict format, locale_t locale, ...),libc_vfw16printf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fw16printf_l,(FILE *__restrict self, char16_t const *__restrict format, locale_t locale, ...),libd_vfw16printf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fw32printf_l,(FILE *__restrict self, char32_t const *__restrict format, locale_t locale, ...),libc_vfw32printf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fw32printf_l,(FILE *__restrict self, char32_t const *__restrict format, locale_t locale, ...),libd_vfw32printf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fprintf_p_l,(FILE *__restrict self, char const *__restrict format, locale_t locale, ...),libc_vfprintf_p_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fprintf_p_l,(FILE *__restrict self, char const *__restrict format, locale_t locale, ...),libd_vfprintf_p_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fw16printf_p_l,(FILE *__restrict self, char16_t const *__restrict format, locale_t locale, ...),libc_vfw16printf_p_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fw16printf_p_l,(FILE *__restrict self, char16_t const *__restrict format, locale_t locale, ...),libd_vfw16printf_p_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fw32printf_p_l,(FILE *__restrict self, char32_t const *__restrict format, locale_t locale, ...),libc_vfw32printf_p_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fw32printf_p_l,(FILE *__restrict self, char32_t const *__restrict format, locale_t locale, ...),libd_vfw32printf_p_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_scanf,(char const *__restrict format, ...),libc_vscanf,(format),format)
DEFINE_PRINTF(ssize_t,libd_scanf,(char const *__restrict format, ...),libd_vscanf,(format),format)
DEFINE_PRINTF(ssize_t,libc_w16scanf,(char16_t const *__restrict format, ...),libc_vw16scanf,(format),format)
DEFINE_PRINTF(ssize_t,libd_w16scanf,(char16_t const *__restrict format, ...),libd_vw16scanf,(format),format)
DEFINE_PRINTF(ssize_t,libc_w32scanf,(char32_t const *__restrict format, ...),libc_vw32scanf,(format),format)
DEFINE_PRINTF(ssize_t,libd_w32scanf,(char32_t const *__restrict format, ...),libd_vw32scanf,(format),format)
DEFINE_PRINTF(ssize_t,libc_scanf_l,(char const *__restrict format, locale_t locale, ...),libc_vscanf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_scanf_l,(char const *__restrict format, locale_t locale, ...),libd_vscanf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w16scanf_l,(char16_t const *__restrict format, locale_t locale, ...),libc_vw16scanf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w16scanf_l,(char16_t const *__restrict format, locale_t locale, ...),libd_vw16scanf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_w32scanf_l,(char32_t const *__restrict format, locale_t locale, ...),libc_vw32scanf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_w32scanf_l,(char32_t const *__restrict format, locale_t locale, ...),libd_vw32scanf_l,(format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fscanf,(FILE *__restrict self, char const *__restrict format, ...),libc_vfscanf,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fscanf,(FILE *__restrict self, char const *__restrict format, ...),libd_vfscanf,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fw16scanf,(FILE *__restrict self, char16_t const *__restrict format, ...),libc_vfw16scanf,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fw16scanf,(FILE *__restrict self, char16_t const *__restrict format, ...),libd_vfw16scanf,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fw32scanf,(FILE *__restrict self, char32_t const *__restrict format, ...),libc_vfw32scanf,(self,format),format)
DEFINE_PRINTF(ssize_t,libd_fw32scanf,(FILE *__restrict self, char32_t const *__restrict format, ...),libd_vfw32scanf,(self,format),format)
DEFINE_PRINTF(ssize_t,libc_fscanf_l,(FILE *__restrict self, char const *__restrict format, locale_t locale, ...),libc_vfscanf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fscanf_l,(FILE *__restrict self, char const *__restrict format, locale_t locale, ...),libd_vfscanf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fw16scanf_l,(FILE *__restrict self, char16_t const *__restrict format, locale_t locale, ...),libc_vfw16scanf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fw16scanf_l,(FILE *__restrict self, char16_t const *__restrict format, locale_t locale, ...),libd_vfw16scanf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libc_fw32scanf_l,(FILE *__restrict self, char32_t const *__restrict format, locale_t locale, ...),libc_vfw32scanf_l,(self,format,locale),locale)
DEFINE_PRINTF(ssize_t,libd_fw32scanf_l,(FILE *__restrict self, char32_t const *__restrict format, locale_t locale, ...),libd_vfw32scanf_l,(self,format,locale),locale)

DECL_END

#endif /* !GUARD_LIBS_LIBC_FORMAT_WRAPPER_C_INL */
