#!/bin/bash

# Build deemon without it already being installed.
#
# Usage:
# $ ./make.sh  [OPTIONS]
#
# OPTIONS may specify a set of variables in the syntax
# `name=value', which can also be passed through environ.
# A list of recognized options can be found below.
#
# Example:
# $ ./make.sh  TARGET_WINDOWS TARGET_386

# Recognized options:
#	ADDR2LINE       = ${CROSS_PREFIX}addr2line
#	AR              = ${CROSS_PREFIX}ar
#	ASM             = ${CROSS_PREFIX}as
#	BUILDPATH       = build/${TARGET_CPU}-kos
#	CC              = ${CROSS_PREFIX}gcc
#	CP              = cp
#	CPP             = ${CROSS_PREFIX}cpp
#	CROSS_PREFIX    = ./binutils/build-binutils-$TARGET_CPU-kos/bin/$TARGET_CPU-kos-
#	CXX             = ${CROSS_PREFIX}g++
#	DISK            = bin/disk-${TARGET_ARCH}-kos.img
#	KERNEL_BINARY   = bin/kernel-${TARGET_CPU}-kos.bin
#	LD              = ${CROSS_PREFIX}gcc
#	MKDIR           = mkdir
#	MTOOLS          = binutils/build-mtools/mtools
#	NM              = ${CROSS_PREFIX}nm
#	OUTPUTPATH      =
#	QEMU            = qemu-system-${TARGET_ARCH}
#	READELF         = ${CROSS_PREFIX}readelf
#	TARGET_ARCH     = i386
#	TARGET_CPU      = i686
#	CONFIG_DEBUG
#	CONFIG_NDEBUG
#	CONFIG_OPTIMIZE

# Automatically generated using:
#	$ deemon -DCONFIG_PRINTCMD magic.dee

function yopt { [[ -v $1 ]]; return $?; }
function nopt { if [[ -v $1 ]]; then return 1; else return 0; fi; }
function iyopt { if [[ -v $1 ]]; then shift; echo $*; fi; }
function inopt { if ! [[ -v $1 ]]; then shift; echo $*; fi; }
function iexpr { if [ $1 ]; then shift; echo $*; fi; }
function run { echo $*; $*; }
function req { $* || exit $?; }
while [[ $# -gt 0 ]]; do
	case $1 in
		-*)
			echo "Unrecognized option: '$1'"
			exit 1
			;;
		*)
			echo "Option: $1"
			if [[ "$1" == *"="* ]]; then
				export $1
			else
				export $1=""
			fi
			;;
	esac
	shift
done
if ! [[ -v OUTPUTPATH ]]; then OUTPUTPATH=""; fi
if ! [[ -v CP ]]; then CP="run cp"; fi
if ! [[ -v MKDIR ]]; then MKDIR="run mkdir"; fi
if ! [[ -v TARGET_CPU ]]; then TARGET_CPU="i686"; fi
if ! [[ -v MTOOLS ]]; then MTOOLS="run binutils/build-mtools/mtools"; fi
if ! [[ -v TARGET_ARCH ]]; then TARGET_ARCH="i386"; fi
if ! [[ -v QEMU ]]; then QEMU="run qemu-system-${TARGET_ARCH}"; fi
if ! [[ -v DISK ]]; then DISK="bin/disk-${TARGET_ARCH}-kos.img"; fi
if ! [[ -v BUILDPATH ]]; then BUILDPATH="build/${TARGET_CPU}-kos"; fi
if ! [[ -v KERNEL_BINARY ]]; then KERNEL_BINARY="bin/kernel-${TARGET_CPU}-kos.bin"; fi
if ! [[ -v CROSS_PREFIX ]]; then CROSS_PREFIX="./binutils/build-binutils-$TARGET_CPU-kos/bin/$TARGET_CPU-kos-"; fi
if ! [[ -v AR ]]; then AR="run ${CROSS_PREFIX}ar"; fi
if ! [[ -v ASM ]]; then ASM="run ${CROSS_PREFIX}as"; fi
if ! [[ -v CXX ]]; then CXX="run ${CROSS_PREFIX}g++"; fi
if ! [[ -v NM ]]; then NM="run ${CROSS_PREFIX}nm"; fi
if ! [[ -v LD ]]; then LD="run ${CROSS_PREFIX}gcc"; fi
if ! [[ -v ADDR2LINE ]]; then ADDR2LINE="run ${CROSS_PREFIX}addr2line"; fi
if ! [[ -v CC ]]; then CC="run ${CROSS_PREFIX}gcc"; fi
if ! [[ -v READELF ]]; then READELF="run ${CROSS_PREFIX}readelf"; fi
if ! [[ -v CPP ]]; then CPP="run ${CROSS_PREFIX}cpp"; fi

${MKDIR} -p ${BUILDPATH}
${MKDIR} -p ${BUILDPATH}/libs.start_debug
${MKDIR} -p ${BUILDPATH}/libs.start
${MKDIR} -p ${BUILDPATH}/libs.libc
${MKDIR} -p ${BUILDPATH}/apps.init
${MKDIR} -p ${BUILDPATH}/kernel.a.out
${MKDIR} -p ${BUILDPATH}/kernel.unix-domain
${MKDIR} -p ${BUILDPATH}/apps.kernctl
${MKDIR} -p ${BUILDPATH}/kernel
${MKDIR} -p ${BUILDPATH}/kernel.ext2fs
${MKDIR} -p ${BUILDPATH}/kernel.pe
${MKDIR} -p ${BUILDPATH}/kernel.procfs
${MKDIR} -p ${BUILDPATH}/apps.terminal-vga
${MKDIR} -p ${BUILDPATH}/libs.libwm
${MKDIR} -p ${BUILDPATH}/apps.wm
${MKDIR} -p ${BUILDPATH}/apps.wms

if nopt CONFIG_NDEBUG; then
CONFIG_DEBUG=1
fi
${MTOOLS} -c mmd -i ${DISK} -D s ::/lib
${MTOOLS} -c mmd -i ${DISK} -D s ::/lib/keymap
req ${MTOOLS} -c mcopy -i ${DISK} -D o disk/de_DE.kmp ::/lib/keymap/de_DE.kmp
req ${MTOOLS} -c mcopy -i ${DISK} -D o disk/en_US.kmp ::/lib/keymap/en_US.kmp

# Project: libs.start_debug
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/libs.start_debug/src.libs.start.start_debug.c.o src/libs/start/start_debug.c
req ${CP} ${BUILDPATH}/libs.start_debug/src.libs.start.start_debug.c.o bin/crt1d.o
${MTOOLS} -c mmd -i ${DISK} -D s ::/usr
${MTOOLS} -c mmd -i ${DISK} -D s ::/usr/lib
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/crt1d.o ::/usr/lib/crt1d.o

# Project: libs.start
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/libs.start/src.libs.start.start.c.o src/libs/start/start.c
req ${CP} ${BUILDPATH}/libs.start/src.libs.start.start.c.o bin/crt1.o
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/crt1.o ::/usr/lib/crt1.o

# Project: libs.libc
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.chtime.c.o src/libs/libc/chtime.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.dirent.c.o src/libs/libc/dirent.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.dl.c.o src/libs/libc/dl.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.dos_crt.c.o src/libs/libc/dos_crt.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.dos_find.c.o src/libs/libc/dos_find.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.dos_slib.c.o src/libs/libc/dos_slib.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.environ.c.o src/libs/libc/environ.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.err.c.o src/libs/libc/err.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.errno.c.o src/libs/libc/errno.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.exec.c.o src/libs/libc/exec.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.exit.c.o src/libs/libc/exit.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.fnmatch.c.o src/libs/libc/fnmatch.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.format-time.c.o src/libs/libc/format-time.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.format.c.o src/libs/libc/format.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.getopt.c.o src/libs/libc/getopt.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.group.c.o src/libs/libc/group.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.heap-trace.c.o src/libs/libc/heap-trace.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.heap.c.o src/libs/libc/heap.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.malloc.c.o src/libs/libc/malloc.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.malloc_d.c.o src/libs/libc/malloc_d.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.malloc_f.c.o src/libs/libc/malloc_f.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.misc.c.o src/libs/libc/misc.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.mount.c.o src/libs/libc/mount.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.posix_signal.c.o src/libs/libc/posix_signal.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.realpath.c.o src/libs/libc/realpath.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.regex.c.o src/libs/libc/regex.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.rtl.c.o src/libs/libc/rtl.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.sched.c.o src/libs/libc/sched.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.stat.c.o src/libs/libc/stat.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.stdio.c.o src/libs/libc/stdio.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.stdio_api.c.o src/libs/libc/stdio_api.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.stdio_wapi.c.o src/libs/libc/stdio_wapi.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.sync.c.o src/libs/libc/sync.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.time.c.o src/libs/libc/time.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.tty.c.o src/libs/libc/tty.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.unicode.c.o src/libs/libc/unicode.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.unistd.c.o src/libs/libc/unistd.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.ushare.c.o src/libs/libc/ushare.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.vm.c.o src/libs/libc/vm.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.widechar.c.o src/libs/libc/widechar.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.stdio.file.c.o src/libs/libc/stdio/file.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.socket.netdb.c.o src/libs/libc/socket/netdb.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.socket.socket.c.o src/libs/libc/socket/socket.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.bound.c.o src/hybrid/bound.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.ctype.c.o src/hybrid/ctype.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.except_errno.c.o src/hybrid/except_errno.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.except_print.c.o src/hybrid/except_print.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.format-printer.c.o src/hybrid/format-printer.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.format-scanner.c.o src/hybrid/format-scanner.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.random.c.o src/hybrid/random.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.string.c.o src/hybrid/string.c
if [[ $TARGET_CPU == i?86 ]]; then
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.entry.c.o src/libs/libc/i386-kos/entry.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.malloc.c.o src/libs/libc/i386-kos/malloc.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall.c.o src/libs/libc/i386-kos/syscall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.entry32.S.o src/libs/libc/i386-kos/entry32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.errno32.S.o src/libs/libc/i386-kos/errno32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.malloc32.S.o src/libs/libc/i386-kos/malloc32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.setjmp32.S.o src/libs/libc/i386-kos/setjmp32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall32.S.o src/libs/libc/i386-kos/syscall32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.except.c.o src/hybrid/i386-kos/except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.instruction.c.o src/hybrid/i386-kos/instruction.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.context32.S.o src/hybrid/i386-kos/context32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.except32.S.o src/hybrid/i386-kos/except32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.string32.S.o src/hybrid/i386-kos/string32.S
fi
if [[ $TARGET_CPU == x86_64 ]]; then
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.entry.c.o src/libs/libc/i386-kos/entry.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.malloc.c.o src/libs/libc/i386-kos/malloc.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall.c.o src/libs/libc/i386-kos/syscall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.errno64.S.o src/libs/libc/i386-kos/errno64.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall64.S.o src/libs/libc/i386-kos/syscall64.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.except.c.o src/hybrid/i386-kos/except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -ffreestanding -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBC -c -o ${BUILDPATH}/libs.libc/src.hybrid.i386-kos.instruction.c.o src/hybrid/i386-kos/instruction.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib "-o bin/libc.so" -static-libgcc -ffreestanding -shared ${BUILDPATH}/libs.libc/src.libs.libc.chtime.c.o ${BUILDPATH}/libs.libc/src.libs.libc.dirent.c.o ${BUILDPATH}/libs.libc/src.libs.libc.dl.c.o ${BUILDPATH}/libs.libc/src.libs.libc.dos_crt.c.o ${BUILDPATH}/libs.libc/src.libs.libc.dos_find.c.o ${BUILDPATH}/libs.libc/src.libs.libc.dos_slib.c.o ${BUILDPATH}/libs.libc/src.libs.libc.environ.c.o ${BUILDPATH}/libs.libc/src.libs.libc.err.c.o ${BUILDPATH}/libs.libc/src.libs.libc.errno.c.o ${BUILDPATH}/libs.libc/src.libs.libc.exec.c.o ${BUILDPATH}/libs.libc/src.libs.libc.exit.c.o ${BUILDPATH}/libs.libc/src.libs.libc.fnmatch.c.o ${BUILDPATH}/libs.libc/src.libs.libc.format-time.c.o ${BUILDPATH}/libs.libc/src.libs.libc.format.c.o ${BUILDPATH}/libs.libc/src.libs.libc.getopt.c.o ${BUILDPATH}/libs.libc/src.libs.libc.group.c.o ${BUILDPATH}/libs.libc/src.libs.libc.heap-trace.c.o ${BUILDPATH}/libs.libc/src.libs.libc.heap.c.o ${BUILDPATH}/libs.libc/src.libs.libc.malloc.c.o ${BUILDPATH}/libs.libc/src.libs.libc.malloc_d.c.o ${BUILDPATH}/libs.libc/src.libs.libc.malloc_f.c.o ${BUILDPATH}/libs.libc/src.libs.libc.misc.c.o ${BUILDPATH}/libs.libc/src.libs.libc.mount.c.o ${BUILDPATH}/libs.libc/src.libs.libc.posix_signal.c.o ${BUILDPATH}/libs.libc/src.libs.libc.realpath.c.o ${BUILDPATH}/libs.libc/src.libs.libc.regex.c.o ${BUILDPATH}/libs.libc/src.libs.libc.rtl.c.o ${BUILDPATH}/libs.libc/src.libs.libc.sched.c.o ${BUILDPATH}/libs.libc/src.libs.libc.stat.c.o ${BUILDPATH}/libs.libc/src.libs.libc.stdio.c.o ${BUILDPATH}/libs.libc/src.libs.libc.stdio_api.c.o ${BUILDPATH}/libs.libc/src.libs.libc.stdio_wapi.c.o ${BUILDPATH}/libs.libc/src.libs.libc.sync.c.o ${BUILDPATH}/libs.libc/src.libs.libc.time.c.o ${BUILDPATH}/libs.libc/src.libs.libc.tty.c.o ${BUILDPATH}/libs.libc/src.libs.libc.unicode.c.o ${BUILDPATH}/libs.libc/src.libs.libc.unistd.c.o ${BUILDPATH}/libs.libc/src.libs.libc.ushare.c.o ${BUILDPATH}/libs.libc/src.libs.libc.vm.c.o ${BUILDPATH}/libs.libc/src.libs.libc.widechar.c.o ${BUILDPATH}/libs.libc/src.libs.libc.stdio.file.c.o ${BUILDPATH}/libs.libc/src.libs.libc.socket.netdb.c.o ${BUILDPATH}/libs.libc/src.libs.libc.socket.socket.c.o ${BUILDPATH}/libs.libc/src.hybrid.bound.c.o ${BUILDPATH}/libs.libc/src.hybrid.ctype.c.o ${BUILDPATH}/libs.libc/src.hybrid.except_errno.c.o ${BUILDPATH}/libs.libc/src.hybrid.except_print.c.o ${BUILDPATH}/libs.libc/src.hybrid.format-printer.c.o ${BUILDPATH}/libs.libc/src.hybrid.format-scanner.c.o ${BUILDPATH}/libs.libc/src.hybrid.random.c.o ${BUILDPATH}/libs.libc/src.hybrid.string.c.o $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.entry.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.malloc.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.entry32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.errno32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.malloc32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.setjmp32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.except.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.instruction.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.context32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.except32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.string32.S.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.entry.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.malloc.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.errno64.S.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.libs.libc.i386-kos.syscall64.S.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.except.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/libs.libc/src.hybrid.i386-kos.instruction.c.o") -lgcc
fi
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/libc.so ::/lib/libc.so

# Project: apps.init
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.init/src.apps.init.main.c.o src/apps/init/main.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib -Lbin -static-libgcc "-o bin/init" ${BUILDPATH}/apps.init/src.apps.init.main.c.o $(iyopt "CONFIG_DEBUG" "bin/crt1d.o") $(inopt "CONFIG_DEBUG" "bin/crt1.o") -lgcc -lc
${MTOOLS} -c mmd -i ${DISK} -D s ::/bin
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/init ::/bin/init

# Project: kernel.a.out
if [[ $TARGET_ARCH == i?86 ]]; then
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.a.out/src.kernel.modules.module.S.o src/kernel/modules/module.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.a.out/src.kernel.modules.a.out.a.out.c.o src/kernel/modules/a.out/a.out.c
req ${CPP} -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -P -D__LINKER__ -o ${BUILDPATH}/kernel.a.out/src.kernel.modules.module.ld.E src/kernel/modules/module.ld
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $([[ $TARGET_ARCH == i?86 ]] && echo "-Wl,-T${BUILDPATH}/kernel.a.out/src.kernel.modules.module.ld.E") $([[ $TARGET_ARCH == i?86 ]] && echo "-shared") $([[ $TARGET_ARCH == i?86 ]] && echo "-Wl,--nmagic") $([[ $TARGET_ARCH == i?86 ]] && echo "-nostdlib") $([[ $TARGET_ARCH == i?86 ]] && echo "-static-libgcc") $([[ $TARGET_ARCH == i?86 ]] && echo "-ffreestanding") $([[ $TARGET_ARCH == i?86 ]] && echo "-o bin/a.out.mod") $([[ $TARGET_ARCH == i?86 ]] && echo "${BUILDPATH}/kernel.a.out/src.kernel.modules.module.S.o") $([[ $TARGET_ARCH == i?86 ]] && echo "${BUILDPATH}/kernel.a.out/src.kernel.modules.a.out.a.out.c.o") $([[ $TARGET_ARCH == i?86 ]] && echo "-lgcc")
${MTOOLS} -c mmd -i ${DISK} -D s ::/mod
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/a.out.mod ::/mod/a.out.mod
fi

# Project: kernel.unix-domain
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.unix-domain/src.kernel.modules.module.S.o src/kernel/modules/module.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.unix-domain/src.kernel.modules.unix-domain.domain.c.o src/kernel/modules/unix-domain/domain.c
req ${CPP} -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -P -D__LINKER__ -o ${BUILDPATH}/kernel.unix-domain/src.kernel.modules.module.ld.E src/kernel/modules/module.ld
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -Wl,-T${BUILDPATH}/kernel.unix-domain/src.kernel.modules.module.ld.E -shared -Wl,--nmagic -nostdlib -static-libgcc -ffreestanding "-o bin/unix-domain.mod" ${BUILDPATH}/kernel.unix-domain/src.kernel.modules.module.S.o ${BUILDPATH}/kernel.unix-domain/src.kernel.modules.unix-domain.domain.c.o -lgcc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/unix-domain.mod ::/mod/unix-domain.mod

# Project: apps.kernctl
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.kernctl/src.apps.kernctl.main.c.o src/apps/kernctl/main.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib -Lbin -static-libgcc "-o bin/kernctl" ${BUILDPATH}/apps.kernctl/src.apps.kernctl.main.c.o $(iyopt "CONFIG_DEBUG" "bin/crt1d.o") $(inopt "CONFIG_DEBUG" "bin/crt1.o") -lgcc -lc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/kernctl ::/bin/kernctl

# Project: kernel
if [[ $TARGET_CPU == i?86 ]]; then
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.except.c.o src/hybrid/i386-kos/except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.instruction.c.o src/hybrid/i386-kos/instruction.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.context32.S.o src/hybrid/i386-kos/context32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.except32.S.o src/hybrid/i386-kos/except32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.string32.S.o src/hybrid/i386-kos/string32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.apic.c.o src/kernel/i386-kos/apic.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.boot.c.o src/kernel/i386-kos/boot.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.cpuid.c.o src/kernel/i386-kos/cpuid.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.debug.c.o src/kernel/i386-kos/debug.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.double_fault.c.o src/kernel/i386-kos/double_fault.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.emulator.c.o src/kernel/i386-kos/emulator.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.except.c.o src/kernel/i386-kos/except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.fpu.c.o src/kernel/i386-kos/fpu.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.gdt.c.o src/kernel/i386-kos/gdt.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.hw_except.c.o src/kernel/i386-kos/hw_except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.illegal_instruction.c.o src/kernel/i386-kos/illegal_instruction.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.init.c.o src/kernel/i386-kos/init.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.interrupt.c.o src/kernel/i386-kos/interrupt.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.ipi.c.o src/kernel/i386-kos/ipi.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.linker.c.o src/kernel/i386-kos/linker.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.memory.c.o src/kernel/i386-kos/memory.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.multiboot.c.o src/kernel/i386-kos/multiboot.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.paging.c.o src/kernel/i386-kos/paging.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.posix_signals.c.o src/kernel/i386-kos/posix_signals.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.rtl.c.o src/kernel/i386-kos/rtl.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.safecall.c.o src/kernel/i386-kos/safecall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.scheduler.c.o src/kernel/i386-kos/scheduler.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.smp.c.o src/kernel/i386-kos/smp.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.syscall.c.o src/kernel/i386-kos/syscall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.sysenter.c.o src/kernel/i386-kos/sysenter.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.task_clone.c.o src/kernel/i386-kos/task_clone.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.tss.c.o src/kernel/i386-kos/tss.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.user.c.o src/kernel/i386-kos/user.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.ushare.c.o src/kernel/i386-kos/ushare.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.vio.c.o src/kernel/i386-kos/vio.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.vm86.c.o src/kernel/i386-kos/vm86.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.vm_kernel.c.o src/kernel/i386-kos/vm_kernel.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.boot32.S.o src/kernel/i386-kos/boot32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.device_interrupts32.S.o src/kernel/i386-kos/device_interrupts32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.hw_except32.S.o src/kernel/i386-kos/hw_except32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.interrupt32.S.o src/kernel/i386-kos/interrupt32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.paging32.S.o src/kernel/i386-kos/paging32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.safecall32.S.o src/kernel/i386-kos/safecall32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.scheduler32.S.o src/kernel/i386-kos/scheduler32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.smp32.S.o src/kernel/i386-kos/smp32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.syscall32.S.o src/kernel/i386-kos/syscall32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.ushare32.S.o src/kernel/i386-kos/ushare32.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.kernel.ld.S.o src/kernel/i386-kos/kernel.ld.S
fi
if [[ $TARGET_CPU == x86_64 ]]; then
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.except.c.o src/hybrid/i386-kos/except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.i386-kos.instruction.c.o src/hybrid/i386-kos/instruction.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.apic.c.o src/kernel/i386-kos/apic.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.boot.c.o src/kernel/i386-kos/boot.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.cpuid.c.o src/kernel/i386-kos/cpuid.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.debug.c.o src/kernel/i386-kos/debug.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.double_fault.c.o src/kernel/i386-kos/double_fault.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.emulator.c.o src/kernel/i386-kos/emulator.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.except.c.o src/kernel/i386-kos/except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.fpu.c.o src/kernel/i386-kos/fpu.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.gdt.c.o src/kernel/i386-kos/gdt.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.hw_except.c.o src/kernel/i386-kos/hw_except.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.illegal_instruction.c.o src/kernel/i386-kos/illegal_instruction.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.init.c.o src/kernel/i386-kos/init.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.interrupt.c.o src/kernel/i386-kos/interrupt.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.ipi.c.o src/kernel/i386-kos/ipi.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.linker.c.o src/kernel/i386-kos/linker.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.memory.c.o src/kernel/i386-kos/memory.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.multiboot.c.o src/kernel/i386-kos/multiboot.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.paging.c.o src/kernel/i386-kos/paging.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.posix_signals.c.o src/kernel/i386-kos/posix_signals.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.rtl.c.o src/kernel/i386-kos/rtl.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.safecall.c.o src/kernel/i386-kos/safecall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.scheduler.c.o src/kernel/i386-kos/scheduler.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.smp.c.o src/kernel/i386-kos/smp.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.syscall.c.o src/kernel/i386-kos/syscall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.sysenter.c.o src/kernel/i386-kos/sysenter.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.task_clone.c.o src/kernel/i386-kos/task_clone.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.tss.c.o src/kernel/i386-kos/tss.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.user.c.o src/kernel/i386-kos/user.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.ushare.c.o src/kernel/i386-kos/ushare.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.vio.c.o src/kernel/i386-kos/vio.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.vm86.c.o src/kernel/i386-kos/vm86.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.vm_kernel.c.o src/kernel/i386-kos/vm_kernel.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.i386-kos.kernel.ld.S.o src/kernel/i386-kos/kernel.ld.S
fi
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.bound.c.o src/hybrid/bound.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.ctype.c.o src/hybrid/ctype.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.except_errno.c.o src/hybrid/except_errno.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.except_print.c.o src/hybrid/except_print.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.format-printer.c.o src/hybrid/format-printer.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.format-scanner.c.o src/hybrid/format-scanner.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.random.c.o src/hybrid/random.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.hybrid.string.c.o src/hybrid/string.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.async_signal.c.o src/kernel/src/kernel/async_signal.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.bind.c.o src/kernel/src/kernel/bind.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.boot.c.o src/kernel/src/kernel/boot.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.cache.c.o src/kernel/src/kernel/cache.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.debug.c.o src/kernel/src/kernel/debug.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.heap.c.o src/kernel/src/kernel/heap.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.ioport.c.o src/kernel/src/kernel/ioport.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.kernctl.c.o src/kernel/src/kernel/kernctl.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.mall.c.o src/kernel/src/kernel/mall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.malloc.c.o src/kernel/src/kernel/malloc.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.meminfo.c.o src/kernel/src/kernel/meminfo.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.memory.c.o src/kernel/src/kernel/memory.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.syscall.c.o src/kernel/src/kernel/syscall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.sysconf.c.o src/kernel/src/kernel/sysconf.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.kernel.system.c.o src/kernel/src/kernel/system.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.unwind.debug_line.c.o src/kernel/src/unwind/debug_line.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.unwind.eh_frame.c.o src/kernel/src/unwind/eh_frame.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.unwind.eh_unwind.c.o src/kernel/src/unwind/eh_unwind.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.unwind.linker.c.o src/kernel/src/unwind/linker.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.affinity.c.o src/kernel/src/sched/affinity.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.group.c.o src/kernel/src/sched/group.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.mutex.c.o src/kernel/src/sched/mutex.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.onexit.c.o src/kernel/src/sched/onexit.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.pid.c.o src/kernel/src/sched/pid.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.posix_signals.c.o src/kernel/src/sched/posix_signals.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.rpc.c.o src/kernel/src/sched/rpc.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.rwlock.c.o src/kernel/src/sched/rwlock.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.semaphore.c.o src/kernel/src/sched/semaphore.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.signal.c.o src/kernel/src/sched/signal.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.suspend.c.o src/kernel/src/sched/suspend.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.task.c.o src/kernel/src/sched/task.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.task_weakref.c.o src/kernel/src/sched/task_weakref.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.unshare.c.o src/kernel/src/sched/unshare.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.userstack.c.o src/kernel/src/sched/userstack.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.sched.userthread.c.o src/kernel/src/sched/userthread.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.core.ata.c.o src/kernel/src/core/ata.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.core.elf.c.o src/kernel/src/core/elf.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.core.fat.c.o src/kernel/src/core/fat.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.net.packetbuffer.c.o src/kernel/src/net/packetbuffer.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.net.socket.c.o src/kernel/src/net/socket.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.net.system.c.o src/kernel/src/net/system.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.corebase.c.o src/kernel/src/vm/corebase.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.environ.c.o src/kernel/src/vm/environ.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.futex.c.o src/kernel/src/vm/futex.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.loadcore.c.o src/kernel/src/vm/loadcore.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.region-io.c.o src/kernel/src/vm/region-io.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.region.c.o src/kernel/src/vm/region.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.system.c.o src/kernel/src/vm/system.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.ushare.c.o src/kernel/src/vm/ushare.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.vm.c.o src/kernel/src/vm/vm.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.vm.ushare.S.o src/kernel/src/vm/ushare.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.keyboard.c.o src/kernel/src/dev/keyboard.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.keymap.c.o src/kernel/src/dev/keymap.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.keymap_default.c.o src/kernel/src/dev/keymap_default.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.mouse.c.o src/kernel/src/dev/mouse.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.pci.c.o src/kernel/src/dev/pci.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.ps2-keyboard.c.o src/kernel/src/dev/ps2-keyboard.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.ps2-mouse.c.o src/kernel/src/dev/ps2-mouse.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.ps2.c.o src/kernel/src/dev/ps2.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.pty.c.o src/kernel/src/dev/pty.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.tty.c.o src/kernel/src/dev/tty.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.dev.wall.c.o src/kernel/src/dev/wall.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.autopart.c.o src/kernel/src/fs/autopart.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.canonbuffer.c.o src/kernel/src/fs/canonbuffer.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.device.c.o src/kernel/src/fs/device.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.driver.c.o src/kernel/src/fs/driver.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.file.c.o src/kernel/src/fs/file.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.handle.c.o src/kernel/src/fs/handle.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.handle_weak.c.o src/kernel/src/fs/handle_weak.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.iov.c.o src/kernel/src/fs/iov.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.linker.c.o src/kernel/src/fs/linker.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.node.c.o src/kernel/src/fs/node.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.path.c.o src/kernel/src/fs/path.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.pipe.c.o src/kernel/src/fs/pipe.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.ramfs.c.o src/kernel/src/fs/ramfs.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.recent_modules.c.o src/kernel/src/fs/recent_modules.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.ringbuffer.c.o src/kernel/src/fs/ringbuffer.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.system.c.o src/kernel/src/fs/system.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.walk.c.o src/kernel/src/fs/walk.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.wordbuffer.c.o src/kernel/src/fs/wordbuffer.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -c -o ${BUILDPATH}/kernel/src.kernel.src.fs.kernel-symtab.S.o src/kernel/src/fs/kernel-symtab.S
req ${CPP} -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -DCONFIG_BUILDING_KERNEL_CORE -P -D__LINKER__ -o ${BUILDPATH}/kernel/src.kernel.$TARGET_ARCH-kos.kernel.ld.E src/kernel/$TARGET_ARCH-kos/kernel.ld
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -Wl,--nmagic -Wl,--eh-frame-hdr -Wl,--no-check-sections -nostdlib "-o $KERNEL_BINARY" -Wl,-T${BUILDPATH}/kernel/src.kernel.$TARGET_ARCH-kos.kernel.ld.E -static-libgcc -ffreestanding $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.except.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.instruction.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.context32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.except32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.string32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.apic.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.boot.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.cpuid.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.debug.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.double_fault.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.emulator.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.except.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.fpu.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.gdt.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.hw_except.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.illegal_instruction.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.init.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.interrupt.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.ipi.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.linker.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.memory.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.multiboot.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.paging.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.posix_signals.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.rtl.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.safecall.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.scheduler.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.smp.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.syscall.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.sysenter.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.task_clone.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.tss.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.user.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.ushare.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.vio.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.vm86.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.vm_kernel.c.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.boot32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.device_interrupts32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.hw_except32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.interrupt32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.paging32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.safecall32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.scheduler32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.smp32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.syscall32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.ushare32.S.o") $([[ $TARGET_CPU == i?86 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.kernel.ld.S.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.except.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.hybrid.i386-kos.instruction.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.apic.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.boot.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.cpuid.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.debug.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.double_fault.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.emulator.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.except.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.fpu.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.gdt.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.hw_except.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.illegal_instruction.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.init.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.interrupt.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.ipi.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.linker.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.memory.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.multiboot.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.paging.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.posix_signals.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.rtl.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.safecall.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.scheduler.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.smp.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.syscall.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.sysenter.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.task_clone.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.tss.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.user.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.ushare.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.vio.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.vm86.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.vm_kernel.c.o") $([[ $TARGET_CPU == x86_64 ]] && echo "${BUILDPATH}/kernel/src.kernel.i386-kos.kernel.ld.S.o") ${BUILDPATH}/kernel/src.hybrid.bound.c.o ${BUILDPATH}/kernel/src.hybrid.ctype.c.o ${BUILDPATH}/kernel/src.hybrid.except_errno.c.o ${BUILDPATH}/kernel/src.hybrid.except_print.c.o ${BUILDPATH}/kernel/src.hybrid.format-printer.c.o ${BUILDPATH}/kernel/src.hybrid.format-scanner.c.o ${BUILDPATH}/kernel/src.hybrid.random.c.o ${BUILDPATH}/kernel/src.hybrid.string.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.async_signal.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.bind.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.boot.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.cache.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.debug.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.heap.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.ioport.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.kernctl.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.mall.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.malloc.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.meminfo.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.memory.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.syscall.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.sysconf.c.o ${BUILDPATH}/kernel/src.kernel.src.kernel.system.c.o ${BUILDPATH}/kernel/src.kernel.src.unwind.debug_line.c.o ${BUILDPATH}/kernel/src.kernel.src.unwind.eh_frame.c.o ${BUILDPATH}/kernel/src.kernel.src.unwind.eh_unwind.c.o ${BUILDPATH}/kernel/src.kernel.src.unwind.linker.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.affinity.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.group.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.mutex.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.onexit.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.pid.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.posix_signals.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.rpc.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.rwlock.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.semaphore.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.signal.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.suspend.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.task.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.task_weakref.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.unshare.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.userstack.c.o ${BUILDPATH}/kernel/src.kernel.src.sched.userthread.c.o ${BUILDPATH}/kernel/src.kernel.src.core.ata.c.o ${BUILDPATH}/kernel/src.kernel.src.core.elf.c.o ${BUILDPATH}/kernel/src.kernel.src.core.fat.c.o ${BUILDPATH}/kernel/src.kernel.src.net.packetbuffer.c.o ${BUILDPATH}/kernel/src.kernel.src.net.socket.c.o ${BUILDPATH}/kernel/src.kernel.src.net.system.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.corebase.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.environ.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.futex.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.loadcore.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.region-io.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.region.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.system.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.ushare.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.vm.c.o ${BUILDPATH}/kernel/src.kernel.src.vm.ushare.S.o ${BUILDPATH}/kernel/src.kernel.src.dev.keyboard.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.keymap.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.keymap_default.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.mouse.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.pci.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.ps2-keyboard.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.ps2-mouse.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.ps2.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.pty.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.tty.c.o ${BUILDPATH}/kernel/src.kernel.src.dev.wall.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.autopart.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.canonbuffer.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.device.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.driver.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.file.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.handle.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.handle_weak.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.iov.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.linker.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.node.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.path.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.pipe.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.ramfs.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.recent_modules.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.ringbuffer.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.system.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.walk.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.wordbuffer.c.o ${BUILDPATH}/kernel/src.kernel.src.fs.kernel-symtab.S.o -lgcc

# Project: kernel.ext2fs
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.ext2fs/src.kernel.modules.module.S.o src/kernel/modules/module.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.ext2fs/src.kernel.modules.ext2fs.ext2.c.o src/kernel/modules/ext2fs/ext2.c
req ${CPP} -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -P -D__LINKER__ -o ${BUILDPATH}/kernel.ext2fs/src.kernel.modules.module.ld.E src/kernel/modules/module.ld
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -Wl,-T${BUILDPATH}/kernel.ext2fs/src.kernel.modules.module.ld.E -shared -Wl,--nmagic -nostdlib -static-libgcc -ffreestanding "-o bin/ext2fs.mod" ${BUILDPATH}/kernel.ext2fs/src.kernel.modules.module.S.o ${BUILDPATH}/kernel.ext2fs/src.kernel.modules.ext2fs.ext2.c.o -lgcc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/ext2fs.mod ::/mod/ext2fs.mod

# Project: kernel.pe
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.pe/src.kernel.modules.module.S.o src/kernel/modules/module.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.pe/src.kernel.modules.pe.loader.c.o src/kernel/modules/pe/loader.c
req ${CPP} -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -P -D__LINKER__ -o ${BUILDPATH}/kernel.pe/src.kernel.modules.module.ld.E src/kernel/modules/module.ld
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -Wl,-T${BUILDPATH}/kernel.pe/src.kernel.modules.module.ld.E -shared -Wl,--nmagic -nostdlib -static-libgcc -ffreestanding "-o bin/pe.mod" ${BUILDPATH}/kernel.pe/src.kernel.modules.module.S.o ${BUILDPATH}/kernel.pe/src.kernel.modules.pe.loader.c.o -lgcc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/pe.mod ::/mod/pe.mod

# Project: kernel.procfs
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.procfs/src.kernel.modules.module.S.o src/kernel/modules/module.S
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.inode.c.o src/kernel/modules/procfs/inode.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.inode_types.c.o src/kernel/modules/procfs/inode_types.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.root.c.o src/kernel/modules/procfs/root.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse2") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-sse4") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-ssse3") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-mmx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-3dnow") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mno-avx") $([[ $TARGET_ARCH == i?86 || $TARGET_ARCH == x86_64 ]] && echo "-mtls-direct-seg-refs") -mpreferred-stack-boundary=2 -shared -fPIC -static-libgcc -ffreestanding -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -c -o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.stringprinter.c.o src/kernel/modules/procfs/stringprinter.c
req ${CPP} -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__KERNEL__ -Isrc/kernel/include -Isrc/kernel/include/$TARGET_ARCH-kos -P -D__LINKER__ -o ${BUILDPATH}/kernel.procfs/src.kernel.modules.module.ld.E src/kernel/modules/module.ld
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -Wl,-T${BUILDPATH}/kernel.procfs/src.kernel.modules.module.ld.E -shared -Wl,--nmagic -nostdlib -static-libgcc -ffreestanding "-o bin/procfs.mod" ${BUILDPATH}/kernel.procfs/src.kernel.modules.module.S.o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.inode.c.o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.inode_types.c.o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.root.c.o ${BUILDPATH}/kernel.procfs/src.kernel.modules.procfs.stringprinter.c.o -lgcc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/procfs.mod ::/mod/procfs.mod

# Project: apps.terminal-vga
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.terminal-vga/src.apps.terminal-vga.libterm.c.o src/apps/terminal-vga/libterm.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.terminal-vga/src.apps.terminal-vga.main.c.o src/apps/terminal-vga/main.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib -Lbin -static-libgcc "-o bin/terminal-vga" ${BUILDPATH}/apps.terminal-vga/src.apps.terminal-vga.libterm.c.o ${BUILDPATH}/apps.terminal-vga/src.apps.terminal-vga.main.c.o $(iyopt "CONFIG_DEBUG" "bin/crt1d.o") $(inopt "CONFIG_DEBUG" "bin/crt1.o") -lgcc -lc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/terminal-vga ::/bin/terminal-vga

# Project: libs.libwm
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.event.c.o src/libs/libwm/event.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.font.c.o src/libs/libwm/font.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.format.c.o src/libs/libwm/format.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.init.c.o src/libs/libwm/init.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.server.c.o src/libs/libwm/server.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.surface.c.o src/libs/libwm/surface.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -shared -fPIC -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -D__BUILDING_LIBWM -c -o ${BUILDPATH}/libs.libwm/src.libs.libwm.window.c.o src/libs/libwm/window.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib "-o bin/libwm.so" -Lbin -static-libgcc -shared ${BUILDPATH}/libs.libwm/src.libs.libwm.event.c.o ${BUILDPATH}/libs.libwm/src.libs.libwm.font.c.o ${BUILDPATH}/libs.libwm/src.libs.libwm.format.c.o ${BUILDPATH}/libs.libwm/src.libs.libwm.init.c.o ${BUILDPATH}/libs.libwm/src.libs.libwm.server.c.o ${BUILDPATH}/libs.libwm/src.libs.libwm.surface.c.o ${BUILDPATH}/libs.libwm/src.libs.libwm.window.c.o -lgcc -lc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/libwm.so ::/lib/libwm.so

# Project: apps.wm
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wm/src.apps.wm.main.c.o src/apps/wm/main.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wm/src.apps.wm.rect.c.o src/apps/wm/rect.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wm/src.apps.wm.window.c.o src/apps/wm/window.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib -Lbin -static-libgcc "-o bin/wm" ${BUILDPATH}/apps.wm/src.apps.wm.main.c.o ${BUILDPATH}/apps.wm/src.apps.wm.rect.c.o ${BUILDPATH}/apps.wm/src.apps.wm.window.c.o $(iyopt "CONFIG_DEBUG" "bin/crt1d.o") $(inopt "CONFIG_DEBUG" "bin/crt1.o") -lgcc -lc -lwm
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/wm ::/bin/wm

# Project: apps.wms
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.display.c.o src/apps/wms/display.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.main.c.o src/apps/wms/main.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.rect.c.o src/apps/wms/rect.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.render.c.o src/apps/wms/render.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.server.c.o src/apps/wms/server.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.window.c.o src/apps/wms/window.c
req ${CC} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") $(inopt "CONFIG_DEBUG" "-DNDEBUG") -Wall -Wextra -Wno-address -Wno-unused-value -Wno-nonnull-compare -Wno-unused-parameter -Wno-comment -Wno-format -Wno-strict-aliasing -Wno-missing-field-initializers -Wno-type-limits -fno-crossjumping -fno-reorder-blocks -fno-reorder-blocks-and-partition -static-libgcc -D__ELF__ -Iinclude -Iinclude/${TARGET_ARCH}-kos -c -o ${BUILDPATH}/apps.wms/src.apps.wms.windowmap.c.o src/apps/wms/windowmap.c
req ${LD} $(iyopt "CONFIG_OPTIMIZE" "-O3") $(iyopt "CONFIG_DEBUG" "-fstack-protector-strong -g") -nostdlib -Lbin -static-libgcc "-o bin/wms" ${BUILDPATH}/apps.wms/src.apps.wms.display.c.o ${BUILDPATH}/apps.wms/src.apps.wms.main.c.o ${BUILDPATH}/apps.wms/src.apps.wms.rect.c.o ${BUILDPATH}/apps.wms/src.apps.wms.render.c.o ${BUILDPATH}/apps.wms/src.apps.wms.server.c.o ${BUILDPATH}/apps.wms/src.apps.wms.window.c.o ${BUILDPATH}/apps.wms/src.apps.wms.windowmap.c.o $(iyopt "CONFIG_DEBUG" "bin/crt1d.o") $(inopt "CONFIG_DEBUG" "bin/crt1.o") -lgcc -lc
req ${MTOOLS} -c mcopy -i ${DISK} -D o bin/wms ::/bin/wms
